# Copyright (c) 2011 Outpost Embedded, LLC
# 
# This file is part of cloudfiles-sftpd.
#
# cloudfiles-sftpd is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# cloudfiles-sftpd is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from cloudfiles_sftpd.cloudfs import READ, WRITE


logger = logging.getLogger(__name__)


###


class AccessManager(object):
    def check_account_permission(self, perm):
        return False

    def check_container_permission(self, container, perm):
        return False

    def check_object_permission(self, obj, perm):
        return False


class AllowAccessManager(object):
    def check_account_permission(self, perm):
        return True

    def check_container_permission(self, container, perm):
        return True

    def check_object_permission(self, obj, perm):
        return True


class SessionPolicyAccessManager(AccessManager):
    policies = None
    session_info = None

    def __init__(self, policies, session_info):
        super(SessionPolicyAccessManager, self).__init__()

        self.policies = policies
        self.session_info = session_info

    def check_account_permission(self, perm):
        for policy in self.policies:
            if policy.check_account_permission(self.session_info, perm):
                return True
        return False

    def check_container_permission(self, container, perm):
        for policy in self.policies:
            if policy.check_container_permission(
              self.session_info,
              container,
              perm,
            ):
                return True
        return False

    def check_object_permission(self, obj, perm):
        for policy in self.policies:
            if policy.check_object_permission(
              self.session_info,
              obj,
              perm,
            ):
                return True
        return False


###


class SessionPolicy(object):
    def check_account_permission(self, session_info, perm):
        return False

    def check_container_permission(self, session_info, container, perm):
        return False

    def check_object_permission(self, session_info, obj, perm):
        return False


class PermissionsPolicy(SessionPolicy):
    perms = None

    def __init__(self, perms):
        super(PermissionsPolicy, self).__init__()

        self.perms = perms

    def check_account_permission(self, session_info, perm):
        return (perm in self.perms)

    def check_container_permission(self, session_info, container, perm):
        return (perm in self.perms)

    def check_object_permission(self, session_info, obj, perm):
        return (perm in self.perms)


class WrapperPolicy(SessionPolicy):
    next_policy = None

    def __init__(self, next_policy):
        super(WrapperPolicy, self).__init__()

        self.next_policy = next_policy


class AccountPolicy(WrapperPolicy):
    def check_account_permission(self, session_info, perm):
        return self.next_policy.check_account_permission(session_info, perm)

    def check_container_permission(self, session_info, container, perm):
        return False

    def check_object_permission(self, session_info, obj, perm):
        return False


class ContainerPolicy(WrapperPolicy):
    container_name = None

    def __init__(self, container_name, next_policy):
        super(ContainerPolicy, self).__init__(next_policy)

        self.container_name = container_name

    def _get_container_name(self, session_info):
        try:
            return self.container_name.format(**session_info)
        except KeyError, e:
            logger.error(unicode(e))
            return None

    def check_account_permission(self, session_info, perm):
        return False

    def check_container_permission(self, session_info, container, perm):
        container_name = self._get_container_name(session_info)
        if container_name is not None and container.name == container_name:
            return self.next_policy.check_container_permission(
              session_info,
              container_name,
              perm,
            )
        return False

    def check_object_permission(self, session_info, obj, perm):
        container_name = self._get_container_name(session_info)
        if container_name is not None and obj.container.name == container_name:
            return self.next_policy.check_object_permission(
              session_info,
              obj,
              perm,
            )
        return False


class AuthMethodPolicy(WrapperPolicy):
    auth_method_name = None

    def __init__(self, auth_method_name, next_policy):
        super(AuthMethodPolicy, self).__init__(next_policy)

        self.auth_method_name = auth_method_name

    def _get_auth_method_name(self, session_info):
        try:
            return session_info['auth_method'].name
        except (KeyError, AttributeError):
            logger.exception('failed to get auth method name')
            return None

    def check_account_permission(self, session_info, perm):
        auth_method_name = self._get_auth_method_name(session_info)
        if (
          auth_method_name is not None
          and auth_method_name == self.auth_method_name
        ):
            return self.next_policy.check_account_permission(
              session_info,
              perm,
            )
        return False

    def check_container_permission(self, session_info, container, perm):
        auth_method_name = self._get_auth_method_name(session_info)
        if (
          auth_method_name is not None
          and auth_method_name == self.auth_method_name
        ):
            return self.next_policy.check_container_permission(
              session_info,
              container,
              perm,
            )
        return False

    def check_object_permission(self, session_info, obj, perm):
        auth_method_name = self._get_auth_method_name(session_info)
        if (
          auth_method_name is not None
          and auth_method_name == self.auth_method_name
        ):
            return self.next_policy.check_object_permission(
              session_info,
              obj,
              perm,
            )
        return False
