# Copyright (c) 2007, 2008  Dwayne C. Litzenberger <dlitz@dlitz.net>
# Copyright (c) 2011 Outpost Embedded, LLC
# 
# This file is part of cloudfiles-sftpd.
#
# cloudfiles-sftpd is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# cloudfiles-sftpd is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from paramiko import AUTH_SUCCESSFUL, AUTH_FAILED

from cloudfiles_sftpd.util import logged


logger = logging.getLogger(__name__)


class AuthManager(object):
    auth_method_types = {}

    @classmethod
    def register_auth_method_type(cls, auth_method_type):
        cls.auth_method_types[auth_method_type.name] = auth_method_type

    ###

    @classmethod
    def ForAuthMethodNames(cls, auth_method_names):
        auth_methods = []
        for auth_method_name in auth_method_names:
            auth_method_type = cls.auth_method_types[auth_method_name]
            auth_method = auth_method_type()
            auth_methods.append(auth_method)
        return cls(auth_methods)

    ###

    auth_methods = None

    def __init__(self, auth_methods):
        super(AuthManager, self).__init__()

        self.auth_methods = auth_methods

    def get_allowed_auths(self, username):
        auths = []
        for auth_method in self.auth_methods:
            for auth in auth_method.get_allowed_auths(username):
                if auth not in auths:
                    auths.append(auth)
        if not auths:
            return 'none'
        return ','.join(auths)

    def _check_auth(self, check, auth_type, username):
        result = None
        for auth_method in self.auth_methods:
            result = check(auth_method)
            if result is not None:
                break
        if result is None:
            auth_method, result = None, AUTH_FAILED
        if result not in (AUTH_SUCCESSFUL, AUTH_FAILED):
            raise AssertionError(result)
        if result == AUTH_FAILED:
            logger.warning(
              'authentication failed: auth_type=%s, username=%s, auth_method=%s',
              auth_type,
              username,
              getattr(auth_method, 'name', '[none]'),
            )
        else:
            logger.info(
              'authentication succeeded: auth_type=%s, username=%s, auth_method=%s',
              auth_type,
              username,
              auth_method.name,
            )
        return auth_method, result

    def check_auth_none(self, username):
        return self._check_auth(
          lambda auth_method: auth_method.check_auth_none(username),
          'none',
          username,
        )

    def check_auth_password(self, username, password):
        return self._check_auth(
          (lambda auth_method:
           auth_method.check_auth_password(username, password)),
          'password',
          username,
        )

    def check_auth_publickey(self, username, key):
        return self._check_auth(
          (lambda auth_method: auth_method.check_auth_publickey(username, key)),
          'publickey',
          username,
        )


class AuthMethod(object):
    def get_allowed_auths(self, username):
        return []

    def check_auth_none(self, username):
        return None

    def check_auth_password(self, username, password):
        return None

    def check_auth_publickey(self, username, key):
        return None


class AnonymousMethod(AuthMethod):
    name = 'anonymous'

    def get_allowed_auths(self, username):
        if username == 'anonymous':
            return ['none']
        return []

    def check_auth_none(self, username):
        if username == 'anonymous':
            return AUTH_SUCCESSFUL
        return None

AuthManager.register_auth_method_type(AnonymousMethod)
