# Copyright (c) 2011 Outpost Embedded, LLC
# 
# This file is part of cloudfiles-sftpd.
#
# cloudfiles-sftpd is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# cloudfiles-sftpd is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Object Representation
=====================


Directories
-----------

Directories are created as described in the Cloud Files Developer Guide.


Symlinks
--------

Symlinks are created similarly to directories.  A symlink is stored with
Content-Type application/symlink with the symlink target written as the file
content.


Unix Permissions
----------------

Permissions are stored on Cloud Files objects as metadata with key "unix-mode",
e.g.:

  >>> obj = filesys.get_instance('/foo/bar/baz')
  >>> obj.metadata['unix-mode'] = '0755'

However, you would normally access it like this:

  >>> print oct(stat.S_IMODE(filesys.lstat('/foo/bar/baz').st_mode))
  0755


Object Access Permissions
=========================

Access permissions are not related to Unix permissions at all.  Rather, Unix
permissions are recorded for archiving purposes only, and SFTP access rights are
granted by an access manager.

Granularity of access permissions is as follows:

Account:

  READ:
    List entries (i.e. list containers).
  WRITE:
    Create entry (i.e. create container).

Container:

  READ:
    List entries (i.e. list objects).  This applies to the container directory
    and any directory within the container.
  WRITE:
    Create entry (i.e. create object).  This applies to the container directory
    and any directory within the container.  Once an object is opened, the
    client has permissions to read and write to it (i.e. permissions for the
    object aren't rechecked after creation).

Object:

  READ:
    Open file for reading, readlink on symlink.  Get attributes.
  WRITE:
    Open file for writing, set attributes.
'''

import os, stat, time, logging, posixpath, rfc822, copy, errno

from cloudfiles import Container, Object
from cloudfiles.errors import (
  NoSuchContainer,
  NoSuchObject,
  ResponseError,
)

from cloudfiles_sftpd.exceptions import OperationError, PermissionDenied
from cloudfiles_sftpd.util import logged


logger = logging.getLogger(__name__)


READ = 'READ'
WRITE = 'WRITE'


DIRECTORY_CONTENT_TYPE = 'application/directory'
SYMLINK_CONTENT_TYPE = 'application/symlink'


class NodeAttributes(object):
    st_size = None
    st_uid = None
    st_gid = None
    st_mode = None
    st_atime = None
    st_mtime = None

    filename = None
    metadata = None

    def __init__(
      self,
      st_size = None,
      st_uid = None,
      st_gid = None,
      st_mode = None,
      st_atime = None,
      st_mtime = None,
      filename = None,
      metadata = None,
    ):
        super(NodeAttributes, self).__init__()

        if metadata is None:
            metadata = {}

        self.st_size = st_size
        self.st_uid = st_uid
        self.st_gid = st_gid
        self.st_mode = st_mode
        self.st_atime = st_atime
        self.st_mtime = st_mtime
        self.filename = filename
        self.metadata = dict(metadata)

    def __repr__(self):
        kwargs = {}
        for name in (
          'st_size',
          'st_uid',
          'st_gid',
          'st_mode',
          'st_atime',
          'st_mtime',
          'filename',
          'metadata',
        ):
            value = getattr(self, name)
            if value is not None:
                kwargs[name] = value
        str_kwargs = ', '.join(
          '%s=%r' % (name, value) for name, value in kwargs.items()
        )
        return '%s(%s)' % (type(self).__name__, str_kwargs)

    ###

    @classmethod
    def ForAccount(cls, insecure = False):
        return cls(st_size = 0, st_mode = (stat.S_IFDIR | 0755))

    @classmethod
    def ForContainer(cls, container, insecure = False):
        return cls(
          st_size = 0,
          st_mode = (stat.S_IFDIR | 0755),
          filename = posixpath.basename(container.name),
        )

    @classmethod
    def ForObject(cls, obj, insecure = False):
        st_mtime = None
        if not insecure:
            # FIXME: Is there a better way to handle this?  It seems buggy for
            # python-cloudfiles to return different date formats.

            # Sometimes we get RFC 822 dates, sometimes we get ISO 8601 dates.
            mtime_tuple = rfc822.parsedate(obj.last_modified)
            if mtime_tuple is None:
                try:
                    # Strip fractional seconds:
                    str_mtime = obj.last_modified.partition('.')[0]
                    # Convert to timestamp:
                    mtime_tuple = time.strptime(str_mtime, '%Y-%m-%dT%H:%M:%S')
                except (ValueError, TypeError, IndexError):
                    pass

            if mtime_tuple is None:
                st_mtime = None
            else:
                st_mtime = time.mktime(mtime_tuple)

        if insecure:
            metadata = {}
        else:
            metadata = dict(obj.metadata)
            if obj.etag:
                metadata['etag'] = obj.etag
            if obj.content_type:
                metadata['content-type'] = obj.content_type

        return cls(
          st_size = obj.size,
          st_mode = cls.get_object_mode(obj, insecure = insecure),
          st_mtime = st_mtime,
          filename = posixpath.basename(obj.name),
          metadata = metadata,
        )

    @classmethod
    def get_object_mode(cls, obj, insecure = False):
        if obj.content_type == SYMLINK_CONTENT_TYPE:
            return (stat.S_IFLNK | stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
        else:
            mode = 0
            if not insecure:
                mode_attr = obj.metadata.get('unix-mode', None)
                if mode_attr is not None:
                    try:
                        mode = int(mode_attr, 8)
                    except ValueError:
                        pass
            if obj.content_type == DIRECTORY_CONTENT_TYPE:
                return mode | stat.S_IFDIR
            else:
                return mode | stat.S_IFREG

    @logged(logger)
    def apply_to_object(self, obj, sync = True):
        # FIXME: We should support setting st_mtime.

        # Attempts to set the following attributes are silently ignored.  Other
        # attributes will be set.
        #
        # * self.st_uid
        # * self.st_gid
        # * self.st_atime
        # * self.st_mtime

        # We *could* support setting st_size, but it would be slow and it
        # doesn't seem particularly useful.  The IETF draft version 13 says:
        #
        #   If this field is present during a setstat operation, the file MUST
        #   be extended or truncated to the specified size.  If the file is
        #   extended, the semantics are to write the byte value 0x00 from the
        #   previous end of file to the new offset.
        #
        # For now, we reject attempts to set it.
        if (self.st_size is not None) and (self.st_size != obj.size):
            raise OperationError(errno.EINVAL)

        metadata = dict(self.metadata)

        obj.content_type = metadata.pop('content-type', obj.content_type)
        obj.etag = metadata.pop('etag', obj.etag)

        if (
          obj.content_type != SYMLINK_CONTENT_TYPE
          and self.st_mode is not None
        ):
            obj.metadata['unix-mode'] = '%04o' % stat.S_IMODE(self.st_mode)

        # Note: If "unix-mode" is in metadata, it will overwrite any value set
        # above.
        obj.metadata.update(metadata)

        if sync:
            obj.sync_metadata()


class BaseObjectNode(object):
    obj = None

    def __init__(self, obj):
        super(BaseObjectNode, self).__init__()

        self.obj = obj

    # FIXME: logging
    def read(self, offset = 0, size = -1):
        raise OperationError(errno.EINVAL)

    # FIXME: logging
    def write(self, data):
        raise OperationError(errno.EINVAL)

    @logged(logger)
    def stat(self):
        # Note: We bypass permission checks enforced by CloudFileSystem.
        return NodeAttributes.ForObject(self.obj)

    @logged(logger)
    def chattr(self, attrs):
        # Note: We bypass permission checks enforced by CloudFileSystem.
        attrs.apply_to_object(self.obj)


class ReadObjectNode(BaseObjectNode):
    def read(self, offset = 0, size = -1):
        try:
            return self.obj.read(size = size, offset = offset)
        except ResponseError, e:
            if e.status == 416:
                # Returning an empty string on EOF is the convention with
                # file-like objects, at least (although we aren't one of
                # those).
                return ''
            raise


class WriteObjectNode(BaseObjectNode):
    def write(self, data):
        if isinstance(data, unicode):
            raise TypeError('data must not be unicode')
        self.obj.write(data)


class ReadWriteObjectNode(ReadObjectNode, WriteObjectNode):
    pass


###


class CloudFileSystem(object):
    conn = None
    access_manager = None
    path_prefix = None

    def __init__(self, conn, access_manager, path_prefix = ''):
        super(CloudFileSystem, self).__init__()

        self.conn = conn
        self.access_manager = access_manager
        self.path_prefix = path_prefix

    ###

    @logged(logger)
    def check_account_permission(self, perm):
        if self.access_manager.check_account_permission(perm):
            return
        logger.warning('DENY %s /', perm)
        raise PermissionDenied()

    @logged(logger)
    def check_container_permission(self, container, perm):
        if self.access_manager.check_container_permission(container, perm):
            return
        logger.warning(
          'DENY %s %s',
          perm,
          self.get_container_path(container, strip_prefix = False),
        )
        raise PermissionDenied()

    @logged(logger)
    def check_object_permission(self, obj, perm):
        if self.access_manager.check_object_permission(obj, perm):
            return
        logger.warning(
          'DENY %s %s',
          perm,
          self.get_object_path(obj, strip_prefix = False),
        )
        raise PermissionDenied()

    ###

    def check_instance_permission(self, instance, perm):
        if instance is None:
            self.check_account_permission(perm)
        elif isinstance(instance, Container):
            self.check_container_permission(instance, perm)
        elif isinstance(instance, Object):
            self.check_object_permission(instance, perm)
        else:
            raise TypeError(instance)

    def check_permission(self, path, perm):
        instance = self.get_instance(path)
        self.check_instance_permission(instance, perm)

    ###

    def _add_prefix(self, path):
        return '%s%s' % (self.path_prefix, path)

    def _strip_prefix(self, path):
        return path[len(self.path_prefix):]

    @logged(logger)
    def parse_path(self, path, add_prefix = True):
        path = self.canonicalize(path)
        if add_prefix:
            path = self._add_prefix(path)

        if not path.startswith('/'):
            raise AssertionError(path)

        if '//' in path:
            raise AssertionError(path)

        # Strip leading slash.
        path = path[1:]

        if path.startswith('/'):
            raise AssertionError(path)

        container_name, slash, subpath = path.partition('/')
        return container_name, subpath

    def get_container_path(self, container, entry = None, strip_prefix = True):
        path = self.path_join('/', container.name)
        if entry is not None:
            path = self.path_join(path, entry)
        if strip_prefix:
            path = self._strip_prefix(path)
        return path

    def get_object_path(self, obj, entry = None, strip_prefix = True):
        path = self.path_join('/', obj.container.name, obj.name)
        if entry is not None:
            path = self.path_join(path, entry)
        if strip_prefix:
            path = self._strip_prefix(path)
        return path

    @logged(logger)
    def get_instance(self, path, add_prefix = True):
        container_name, subpath = self.parse_path(path, add_prefix = add_prefix)
        if not container_name and not subpath:
            return None

        try:
            container = self.conn.get_container(container_name)
        except NoSuchContainer:
            # Don't reveal unauthorized information.
            self.check_account_permission(READ)
            raise

        if not subpath:
            return container

        try:
            obj = container.get_object(subpath)
        except NoSuchObject:
            # Don't reveal unauthorized information.
            self.check_container_permission(container, READ)
            raise

        return obj

    @logged(logger)
    def create_instance(
      self,
      path,
      attrs = None,
      default_mode = 0644,
      container_allowed = False,
      add_prefix = True,
    ):
        container_name, subpath = self.parse_path(path, add_prefix = add_prefix)

        if not container_name:
            raise OperationError(errno.EINVAL)

        if not subpath:
            if not container_allowed:
                # Note: This is not a permission check.  This prevents creating
                # symlinks and regular files in the root directory.  Only
                # directories (containers) can be created there.
                raise OperationError(errno.EINVAL)

            try:
                container = self.conn.get_container(container_name)
            except NoSuchContainer:
                pass
            else:
                # Don't reveal unauthorized information.
                try:
                    self.check_account_permission(READ)
                except PermissionDenied:
                    self.check_container_permission(container, READ)
                raise OperationError(errno.EEXIST)

            self.check_account_permission(WRITE)
            return self.conn.create_container(container_name)

        try:
            container = self.conn.get_container(container_name)
        except NoSuchContainer:
            self.check_account_permission(READ)
            raise

        try:
            obj = container.get_object(subpath)
        except NoSuchObject:
            pass
        else:
            try:
                self.check_container_permission(container, READ)
            except PermissionDenied:
                self.check_object_permission(obj, READ)
            raise OperationError(errno.EEXIST)

        self.check_container_permission(container, WRITE)

        parent_subpath = posixpath.dirname(subpath)
        if parent_subpath:
            parent_instance = container.get_object(parent_subpath)
            if not stat.S_ISDIR(NodeAttributes.get_object_mode(parent_instance)):
                # Don't reveal unauthorized information.
                self.check_object_permission(parent_instance, READ)
                raise OperationError(errno.ENOTDIR)

        instance = container.create_object(subpath)
        if attrs is None:
            attrs = NodeAttributes()
        # Note: st_mode is never applied to symlinks (see apply_to_object).
        if attrs.st_mode is None:
            attrs = copy.copy(attrs)
            attrs.st_mode = default_mode
        attrs.apply_to_object(instance, sync = False)
        return instance

    def require_object_or_container(self, instance):
        if not isinstance(instance, (Object, Container)):
            # Don't reveal unauthorized information.
            self.check_instance_permission(instance, READ)
            raise OperationError(errno.EINVAL)

    def require_object(self, instance):
        if not isinstance(instance, Object):
            # Don't reveal unauthorized information.
            self.check_instance_permission(instance, READ)
            raise OperationError(errno.EINVAL)

    def path_join(self, *p):
        return posixpath.join(*p)

    ###

    @logged(logger)
    def open(self, path, flags, attrs = None):
        if flags & ~(os.O_CREAT | os.O_EXCL | os.O_WRONLY | os.O_RDWR):
            raise ValueError(
              'only O_CREAT, O_EXCL, O_WRONLY, O_RDWR are supported'
            )

        try:
            instance = self.get_instance(path)
        except (PermissionDenied, NoSuchContainer, NoSuchObject), e:
            if not (flags & os.O_CREAT):
                # Note: It's important that we preserve the exception type
                # because we might not have sufficient permissions to see
                # NoSuch{Container,Object}.
                raise
            else:
                instance = self.create_instance(path, attrs, default_mode = 0644)
                # The object should exist after we open it, even if no data is
                # written.
                instance.write()
        else:
            self.require_object(instance)
            mode = NodeAttributes.get_object_mode(instance)
            if stat.S_ISDIR(mode):
                # Don't reveal unauthorized information.
                self.check_object_permission(instance, READ)
                raise OperationError(errno.EISDIR)
            if stat.S_ISLNK(mode):
                # Don't reveal unauthorized information.
                self.check_object_permission(instance, READ)
                # We don't dereference symlinks.
                # FIXME: Is EINVAL the best error to use here?
                raise OperationError(errno.EINVAL)
            if (flags & os.O_CREAT) and (flags & os.O_EXCL):
                # Don't reveal unauthorized information.
                try:
                    self.check_container_permission(instance.container, READ)
                except PermissionDenied:
                    self.check_object_permission(instance, READ)
                raise OperationError(errno.EEXIST)

            # Note: We only check object permissions for previously existing
            # objects.  If we had sufficient permissions to create an object,
            # we assume that is all we need to be allowed to work with it.

            if not (flags & os.O_WRONLY):
                self.check_object_permission(instance, READ)

            if flags & (os.O_WRONLY | os.O_RDWR):
                self.check_object_permission(instance, WRITE)

        if flags & os.O_WRONLY:
            cls = WriteObjectNode
        elif flags & os.O_RDWR:
            cls = ReadWriteObjectNode
        else:
            cls = ReadObjectNode

        return cls(instance)

    @logged(logger)
    def listdir(self, path):
        instance = self.get_instance(path)

        if instance is None:
            self.check_account_permission(READ)
            result = []
            for container in self.conn.get_all_containers():
                try:
                    self.check_container_permission(container, READ)
                except PermissionDenied:
                    insecure = True
                else:
                    insecure = False
                result.append(
                  NodeAttributes.ForContainer(container, insecure = insecure)
                )
            return result

        if isinstance(instance, Container):
            container = instance
        elif isinstance(instance, Object):
            container = instance.container
        else:
            raise AssertionError(instance)

        self.check_container_permission(container, READ)

        if isinstance(instance, Container):
            objects = [
              instance.get_object(name)
              for name in instance.list_objects(path = '')
            ]
        else:
            if not stat.S_ISDIR(NodeAttributes.get_object_mode(instance)):
                # Don't reveal unauthorized information.
                self.check_object_permission(instance, READ)
                raise OperationError(errno.EINVAL)
            objects = [
              instance.container.get_object(name)
              for name in instance.container.list_objects(path = instance.name)
            ]

        result = []
        for obj in objects:
            # If we can't read the object, we should not give out information
            # about it (size, type, etag, metadata, etc.).
            try:
                self.check_object_permission(obj, READ)
            except PermissionDenied:
                insecure = True
            else:
                insecure = False
            result.append(NodeAttributes.ForObject(obj, insecure = insecure))
        return result

    @logged(logger)
    def stat(self, path):
        return self.lstat(path)

    @logged(logger)
    def lstat(self, path):
        instance = self.get_instance(path)
        self.check_instance_permission(instance, READ)
        if instance is None:
            return NodeAttributes.ForAccount()
        if isinstance(instance, Container):
            return NodeAttributes.ForContainer(instance)
        return NodeAttributes.ForObject(instance)

    @logged(logger)
    def remove(self, path):
        instance = self.get_instance(path)
        self.require_object(instance)
        if stat.S_ISDIR(NodeAttributes.get_object_mode(instance)):
            # Don't reveal unauthorized information.
            self.check_object_permission(instance, READ)
            raise OperationError(errno.EINVAL)
        self.check_container_permission(instance.container, WRITE)
        instance.container.delete_object(instance.name)

    def _copy(self, oldpath, newpath, rename = False):
        old_instance = self.get_instance(oldpath)
        # Copying containers is not supported.
        self.require_object(old_instance)
        self.check_object_permission(old_instance, READ)

        new_instance = self.create_instance(newpath)
        self.require_object(new_instance)

        # FIXME: In a future version of python-cloudfiles, this may not be
        # necessary.

        # It appears that we have to copy the etag, content type and metadata
        # to the new instance ourselves.  The Cloud Files service does not copy
        # these attributes itself, and the python-cloudfiles copy
        # implementation doesn't do it for us, either.

        new_instance.etag = old_instance.etag
        new_instance.content_type = old_instance.content_type
        new_instance.metadata.update(old_instance.metadata)

        new_instance.copy_from(old_instance.container.name, old_instance.name)

        if stat.S_ISDIR(NodeAttributes.get_object_mode(old_instance)):
            for old_child_subpath in old_instance.container.list_objects(
              path = old_instance.name
            ):
                entry = posixpath.basename(old_child_subpath)
                old_child_path = self.get_object_path(old_instance, entry)
                new_child_path = self.get_object_path(new_instance, entry)
                self._copy(old_child_path, new_child_path, rename = rename)

        if rename:
            old_instance.container.delete_object(old_instance.name)

        return new_instance

    @logged(logger)
    def copy(self, oldpath, newpath):
        # FIXME: This copies directories recursively by default.
        return self._copy(oldpath, newpath, rename = False)

    @logged(logger)
    def rename(self, oldpath, newpath):
        return self._copy(oldpath, newpath, rename = True)

    @logged(logger)
    def mkdir(self, path, attrs = None):
        instance = self.create_instance(
          path,
          attrs,
          default_mode = 0755,
          container_allowed = True,
        )
        if isinstance(instance, Object):
            instance.content_type = DIRECTORY_CONTENT_TYPE
            instance.write()
        return instance

    @logged(logger)
    def rmdir(self, path):
        instance = self.get_instance(path)

        self.require_object_or_container(instance)

        if isinstance(instance, Container):
            if instance.list_objects():
                # Don't reveal unauthorized information.
                self.check_container_permission(instance, READ)
                raise OperationError(errno.ENOTEMPTY)
            self.check_account_permission(WRITE)
            self.conn.delete_container(instance.name)
            return

        if not stat.S_ISDIR(NodeAttributes.get_object_mode(instance)):
            # Don't reveal unauthorized information.
            self.check_object_permission(instance, READ)
            raise OperationError(errno.ENOTDIR)

        if instance.container.list_objects(path = instance.name):
            # Don't reveal unauthorized information.
            self.check_object_permission(instance, READ)
            raise OperationError(errno.ENOTEMPTY)

        self.check_container_permission(instance.container, WRITE)
        instance.container.delete_object(instance.name)

    @logged(logger)
    def chattr(self, path, attrs):
        instance = self.get_instance(path)
        self.require_object(instance)
        self.check_object_permission(instance, WRITE)
        attrs.apply_to_object(instance)

    @logged(logger)
    def canonicalize(self, path):
        if posixpath.isabs(path):
            return posixpath.normpath(path)
        else:
            return posixpath.normpath('/%s' % path)

    @logged(logger)
    def readlink(self, path):
        instance = self.get_instance(path)
        self.require_object(instance)
        self.check_object_permission(instance, READ)
        if not stat.S_ISLNK(NodeAttributes.get_object_mode(instance)):
            raise OperationError(errno.EINVAL)
        return instance.read()

    @logged(logger)
    def symlink(self, target_path, path):
        instance = self.create_instance(path)
        instance.content_type = SYMLINK_CONTENT_TYPE
        instance.write(target_path)
        return instance

    ###

    chunk_size = 1000000

    def fetch_dir(self, remote_path, local_path, use_existing = False):
        # FIXME: We should optionally copy user, group, and permissions.

        try:
            os.mkdir(local_path)
        except OSError, e:
            if e.errno != errno.EEXIST:
                raise
            if not use_existing:
                raise
        for attrs in self.listdir(remote_path):
            local_entry_path = os.path.join(local_path, attrs.filename)
            remote_entry_path = self.path_join(remote_path, attrs.filename)
            if stat.S_ISDIR(attrs.st_mode):
                self.fetch_dir(remote_entry_path, local_entry_path)
            elif stat.S_ISLNK(attrs.st_mode):
                self.fetch_symlink(remote_entry_path, local_entry_path)
            else:
                self.fetch_file(remote_entry_path, local_entry_path)

    def fetch_symlink(self, remote_path, local_path):
        os.symlink(self.readlink(remote_path), local_path)

    def fetch_file(self, remote_path, local_path):
        # FIXME: We should optionally copy user, group, and permissions.

        offset = 0
        with open(local_path, 'wb') as f:
            node = self.open(remote_path, os.O_RDONLY)
            while True:
                chunk = node.read(offset, self.chunk_size)
                if not chunk:
                    break
                f.write(chunk)
                offset += len(chunk)
