# Copyright (c) 2011 Outpost Embedded, LLC
# 
# This file is part of cloudfiles-sftpd.
#
# cloudfiles-sftpd is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# cloudfiles-sftpd is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from ConfigParser import RawConfigParser, NoSectionError, NoOptionError

from paramiko import (
  SSHException,
  RSAKey,
  DSSKey,
)

from cloudfiles_sftpd.access import (
  PermissionsPolicy,
  AccountPolicy,
  ContainerPolicy,
  AuthMethodPolicy,
  READ,
  WRITE,
)
from cloudfiles_sftpd.exceptions import ConfigurationError


logger = logging.getLogger(__name__)


class CloudFilesSftpdConfigParser(RawConfigParser):
    def get_log(self):
        try:
            return self.get('global', 'log')
        except (NoSectionError, NoOptionError):
            return '-'

    def get_log_level(self):
        try:
            return self.get('global', 'log_level')
        except (NoSectionError, NoOptionError):
            return 'info'

    def get_username(self):
        return self.get('global', 'username')

    def get_api_key(self):
        return self.get('global', 'api_key')

    def get_cloud_files_timeout(self):
        try:
            return self.getint('global', 'cloud_files_timeout')
        except (NoSectionError, NoOptionError, ValueError):
            return None

    def get_addr(self):
        try:
            s = self.get('global', 'listen')
        except (NoSectionError, NoOptionError):
            return '0.0.0.0', 22
        if ':' in s:
            host, port = s.split(':')
            port = int(port)
            return host, port
        else:
            return s, 22

    def get_host_keys(self):
        try:
            str_host_keys = self.get('global', 'host_keys')
        except (NoSectionError, NoOptionError):
            return []
        host_key_paths = [s.strip() for s in str_host_keys.split(',')]
        host_keys = []
        for filename in host_key_paths:
            try:
                host_key = RSAKey.from_private_key_file(filename = filename)
            except SSHException:
                host_key = DSSKey.from_private_key_file(filename = filename)
            host_keys.append(host_key)
            del host_key
        return host_keys

    def get_keepalive(self):
        try:
            keepalive = self.getint('global', 'keepalive')
        except (NoSectionError, NoOptionError, ValueError):
            keepalive = None
        if keepalive is not None:
            if keepalive <= 0:
                keepalive = None
        return keepalive

    def _option_is_permission(self, option_name):
        return ((option_name == '*') or option_name.startswith('/'))

    def get_auth_method_names(self):
        auth_method_names = list(self.sections())
        auth_method_names.remove('global')
        return auth_method_names

    def get_access_policies(self):
        policies = []

        for auth_method_name in self.get_auth_method_names():
            for option_name in self.options(auth_method_name):
                if not self._option_is_permission(option_name):
                    continue

                value = self.get(auth_method_name, option_name)

                logger.debug(
                  'found policy for %s: %s = %s',
                  auth_method_name,
                  option_name,
                  value,
                )

                if value == 'ro':
                    perms = [READ]
                elif value == 'wo':
                    perms = [WRITE]
                elif value == 'rw':
                    perms = [READ, WRITE]
                else:
                    raise ConfigurationError(
                      'permissions must be specified as "ro", "wo", or "rw": %s'
                      % option_name
                    )

                policy = PermissionsPolicy(perms)

                if option_name == '*':
                    pass
                elif option_name == '/':
                    policy = AccountPolicy(policy)
                else:
                    if option_name.count('/') != 1:
                        raise ConfigurationError(
                          'granting permissions for non-account, '
                          'non-container paths is not supported: %s'
                          % option_name
                        )
                    container_name = option_name[1:]
                    policy = ContainerPolicy(container_name, policy)

                policy = AuthMethodPolicy(auth_method_name, policy)
                policies.append(policy)

        return policies

    def get_auth_method_options(self):
        auth_method_options = {}
        for auth_method_name in self.get_auth_method_names():
            d = {}
            auth_method_options[auth_method_name] = d
            for option_name in self.options(auth_method_name):
                if self._option_is_permission(option_name):
                    continue
                # FIXME: Generalize handling of option types.
                if option_name in ('lock_container', 'create_root'):
                    value = self.getboolean(auth_method_name, option_name)
                else:
                    value = self.get(auth_method_name, option_name)
                d[option_name] = value
        return auth_method_options

    def get_extensions(self):
        try:
            str_extensions = self.get('global', 'extensions')
        except (NoSectionError, NoOptionError):
            return []
        return [s.strip() for s in str_extensions.split(',')]
