# Copyright (c) 2007, 2008  Dwayne C. Litzenberger <dlitz@dlitz.net>
# Copyright (c) 2011 Outpost Embedded, LLC
# 
# This file is part of cloudfiles-sftpd.
#
# cloudfiles-sftpd is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# cloudfiles-sftpd is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from SocketServer import (
  TCPServer,
  ThreadingMixIn,
  BaseRequestHandler,
)

import paramiko
from cloudfiles import ConnectionPool

from cloudfiles_sftpd.server_interface import CloudFilesServerInterface
from cloudfiles_sftpd.auth import AuthManager
from cloudfiles_sftpd.access import SessionPolicyAccessManager
from cloudfiles_sftpd.sftp_server_interface import CloudFilesSFTPServerInterface
from cloudfiles_sftpd.exceptions import ProtocolError


logger = logging.getLogger(__name__)


class SFTPServer(ThreadingMixIn, TCPServer):
    @classmethod
    def ForConfig(cls, config):
        kwargs = {
          'username': config.get_username(), 
          'api_key': config.get_api_key(),
        }
        timeout = config.get_cloud_files_timeout()
        # If no timeout was specified in the configuration, we don't
        # specify one either.  The python-cloudfiles default will be
        # used in that case.
        if timeout is not None:
            kwargs['timeout'] = timeout
        connpool = ConnectionPool(**kwargs)

        return cls(
          auth_manager = AuthManager.ForAuthMethodNames(
            config.get_auth_method_names(),
          ),
          connpool = connpool,
          access_policies = config.get_access_policies(),
          auth_method_options = config.get_auth_method_options(),
          addr = config.get_addr(),
          host_keys = config.get_host_keys(),
          keepalive = config.get_keepalive(),
        )

    ###

    auth_manager = None
    connpool = None
    access_policies = None
    auth_method_options = None
    host_keys = None
    keepalive = None

    transports = None

    def __init__(
      self,
      auth_manager,
      connpool,
      access_policies = (),
      auth_method_options = {},
      addr = ('', 22),
      host_keys = (),
      keepalive = None,
    ):
        TCPServer.__init__(self, addr, SFTPConnectionRequestHandler)

        self.auth_manager = auth_manager
        self.connpool = connpool
        self.access_policies = access_policies
        self.auth_method_options = auth_method_options
        self.host_keys = host_keys
        self.keepalive = keepalive

        self.transports = []

    ###

    def add_transport(self, transport):
        logger.debug('adding transport %r', transport)
        self.transports.append(transport)

    def remove_transport(self, transport):
        logger.debug('removing transport %r', transport)
        try:
            self.transports.remove(transport)
        except ValueError:
            logger.debug('failed to remove unknown transport: %r', transport)

    def close_transports(self):
        logger.info('closing open sessions...')
        closed_transports = []
        while self.transports:
            transport = self.transports.pop()
            logger.debug('closing transport %r', transport)
            transport.close()
            closed_transports.append(transport)
        for transport in closed_transports:
            transport.join()
        logger.info('closed %u open sessions', len(closed_transports))


class SFTPConnectionRequestHandler(BaseRequestHandler):
    auth_timeout = 120
    cloudfiles_conn = None

    def setup(self):
        self.transport = self.make_transport(self.request)
        self.server.add_transport(self.transport)
        if self.server.keepalive is not None:
            self.transport.set_keepalive(self.server.keepalive)
        self.load_server_moduli()
        self.set_security_options()
        self.add_host_keys()
        self.set_subsystem_handlers()

    def handle(self):
        self.transport.start_server(
          server = CloudFilesServerInterface(self.server.auth_manager),
        )
        # Get the session channel
        chan = self.transport.accept(self.auth_timeout)
        if chan is None:
            raise ProtocolError(
              'session channel not opened (authentication failure?)'
            )
        self.transport.join()
        self.server.remove_transport(self.transport)

    def make_transport(self, sock):
        return paramiko.Transport(sock)

    def load_server_moduli(self):
        self.transport.load_server_moduli()

    def set_security_options(self):
        so = self.transport.get_security_options()

        # Don't support any of (hmac-sha1-96, hmac-md5, hmac-md5-96, none)
        so.digests = ('hmac-sha1',)
   
        # Support delayed zlib compression, but not 'zlib' compression.
        # 'zlib@openssh.com' does the same thing as 'zlib', but avoids attacks
        # by unauthenticated users.
        so.compression = ('zlib@openssh.com', 'none')
        
    def add_host_keys(self):
        for key in self.server.host_keys:
            self.transport.add_server_key(key)

    def set_subsystem_handlers(self):
        self.transport.set_subsystem_handler(
          'sftp',
          paramiko.SFTPServer,
          sftp_si = CloudFilesSFTPServerInterface,
          transport = self.transport,
          connpool = self.server.connpool,
          access_policies = self.server.access_policies,
          auth_method_options = self.server.auth_method_options,
        )
