# Copyright (c) 2007, 2008  Dwayne C. Litzenberger <dlitz@dlitz.net>
# Copyright (c) 2011 Outpost Embedded, LLC
# 
# This file is part of cloudfiles-sftpd.
#
# cloudfiles-sftpd is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# cloudfiles-sftpd is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, stat, time, logging, posixpath, rfc822, errno
from functools import wraps
from tempfile import TemporaryFile
from Queue import Queue
from threading import Thread

from cloudfiles.errors import (
  NoSuchContainer,
  NoSuchObject,
  ResponseError,
)
from paramiko import (
  SFTPServerInterface,
  SFTPAttributes,
  SFTP_PERMISSION_DENIED,
  SFTP_NO_SUCH_FILE,
  SFTP_EOF,
  SFTP_FAILURE,
  SFTP_OP_UNSUPPORTED,
  SFTP_OK,
  SFTPHandle,
)

from cloudfiles_sftpd.access import AllowAccessManager, SessionPolicyAccessManager
from cloudfiles_sftpd.cloudfs import CloudFileSystem, NodeAttributes
from cloudfiles_sftpd.locking import lock_container, unlock_container
from cloudfiles_sftpd.exceptions import ConfigurationError, OperationError
from cloudfiles_sftpd.util import logged


logger = logging.getLogger(__name__)


###


ERRNO_TO_SFTP = {
  errno.ENOENT: SFTP_NO_SUCH_FILE,
  errno.EACCES: SFTP_PERMISSION_DENIED,
  errno.EISDIR: SFTP_FAILURE,
  errno.ENOTDIR: SFTP_FAILURE,
  errno.ENOTEMPTY: SFTP_FAILURE,
  errno.EEXIST: SFTP_FAILURE,
  errno.EINVAL: SFTP_FAILURE,
}


def convert_fs_exceptions(fn):
    @wraps(fn)
    def wrapped(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except (NoSuchContainer, NoSuchObject):
            return SFTP_NO_SUCH_FILE
        except OperationError, e:
            return ERRNO_TO_SFTP.get(e.errno, SFTP_FAILURE)
    return wrapped


###


def file_is_zero_length(truncate, object_node):
    zero_length = False
    if truncate:
        zero_length = True
    if object_node.obj.size in (0, None):
        zero_length = True
    return zero_length


@logged(logger)
def make_handle(object_node, flags):
    truncate = bool(flags & os.O_TRUNC)
    append_only = bool(flags & os.O_APPEND)

    zero_length = file_is_zero_length(truncate, object_node)

    if flags & os.O_WRONLY:
        if zero_length:
            if append_only:
                cls = StreamingWriteObjectHandle
            else:
                cls = OptimisticStreamingWriteObjectHandle
        else:
            cls = WriteObjectHandle
    elif flags & os.O_RDWR:
        cls = ReadWriteObjectHandle
    else:
        cls = ReadObjectHandle

    return cls(object_node, truncate, append_only)


class BaseObjectHandle(SFTPHandle):
    object_node = None
    truncate = None
    append_only = None

    def __init__(self, object_node, truncate, append_only):
        # Note: We don't call SFTPHandle.__init__ because it only sets up for
        # the base class implementation, which we don't use.
        #super(BaseObjectHandle, self).__init__()

        self.object_node = object_node
        self.truncate = truncate
        self.append_only = append_only

    def close(self):
        return SFTP_OK

    def read(self, offset, length):
        return SFTP_OP_UNSUPPORTED

    def write(self, offset, data):
        return SFTP_OP_UNSUPPORTED

    @convert_fs_exceptions
    def stat(self):
        return node_attrs_to_sftp_attrs(self.object_node.stat())

    @convert_fs_exceptions
    @logged(logger)
    def chattr(self, attr):
        self.object_node.chattr(sftp_attrs_to_node_attrs(attr))
        return SFTP_OK


class ReadObjectHandle(BaseObjectHandle):
    def __init__(self, object_node, truncate, append_only):
        if truncate:
            raise TypeError('ReadObjectHandle: truncate must be False')
        if append_only:
            raise TypeError('ReadObjectHandle: append_only must be False')

        super(ReadObjectHandle, self).__init__(
          object_node,
          truncate,
          append_only,
        )

    @convert_fs_exceptions
    def read(self, offset, length):
        data = self.object_node.read(offset = offset, size = length)
        if not data:
            return SFTP_EOF
        return data


class WriteObjectHandle(BaseObjectHandle):
    def __init__(self, object_node, truncate, append_only):
        super(WriteObjectHandle, self).__init__(
          object_node,
          truncate,
          append_only,
        )

        self.fobj = TemporaryFile()
        if not truncate:
            self.object_node.obj.read(buffer = self.fobj)

    @convert_fs_exceptions
    def close(self):
        # FIXME: We only need to push changes to Cloud Files if data was
        # actually written.
        self.fobj.seek(0)
        # FIXME: Errors here don't get passed along to the SFTP client.
        self.object_node.obj.write(self.fobj)
        self.fobj.close()
        return SFTP_OK

    def write(self, offset, data):
        if self.append_only:
            # Seek to end of file.
            self.fobj.seek(0, 2)
        else:
            self.fobj.seek(offset)
        self.fobj.write(data)
        return SFTP_OK

    def stat(self):
        attrs = super(WriteObjectHandle, self).stat()
        if not isinstance(attrs, SFTPAttributes):
            return attrs
        # FIXME: Should we update other parameters (e.g. atime, mtime)?
        self.fobj.seek(0, 2)
        attrs.st_size = self.fobj.tell()
        return attrs


class ReadWriteObjectHandle(WriteObjectHandle):
    def read(self, offset, length):
        self.fobj.seek(offset)
        data = self.fobj.read(length)
        if not data:
            return SFTP_EOF
        return data


class StreamingMixin(object):
    _queue = None
    _streaming_thread = None

    def _send(self, data):
        self._queue.put(data)

    def _start_streaming_thread(self):
        self._queue = Queue()
        self._streaming_thread = Thread(target = self._stream)
        self._streaming_thread.start()

    def _stop_streaming_thread(self):
        self._send(None)
        self._streaming_thread.join()
        self._streaming_thread = None
        self._queue = None

    def _stream(self):
        # Fetch a fresh copy of the storage object to avoid problems if someone
        # else is holding a reference to this one.
        obj = self.object_node.obj.container.get_object(
          self.object_node.obj.name,
        )
        # We have to clear out the object size because otherwise it will be
        # sent as the Content-Length header (which is obviously not what we
        # want).
        obj.size = None
        # FIXME: Errors here don't get passed along to the SFTP client.
        obj.send(self._iter_data())

    def _iter_data(self):
        while True:
            data = self._queue.get()
            if data is None:
                break
            yield data


class StreamingWriteObjectHandle(StreamingMixin, BaseObjectHandle):
    def __init__(self, object_node, truncate, append_only):
        if not append_only:
            raise AssertionError('append_only must be True')

        if not file_is_zero_length(truncate, object_node):
            raise AssertionError('truncate = %r, size = %r' % (
              truncate,
              object_node.obj.size,
            ))

        super(StreamingWriteObjectHandle, self).__init__(
          object_node,
          truncate,
          append_only,
        )

        self._start_streaming_thread()

    def close(self):
        self._stop_streaming_thread()
        return super(StreamingWriteObjectHandle, self).close()

    def write(self, offset, data):
        # Note: We completely ignore offset (assume O_APPEND).
        self._send(data)
        return SFTP_OK


class OptimisticStreamingWriteObjectHandle(
  StreamingMixin,
  WriteObjectHandle,
):
    _next_offset = None

    def __init__(self, object_node, truncate, append_only):
        if append_only:
            raise AssertionError('append_only must be False')

        if not file_is_zero_length(truncate, object_node):
            raise AssertionError('truncate = %r, size = %r' % (
              truncate,
              object_node.obj.size,
            ))

        self._next_offset = 0

        super(OptimisticStreamingWriteObjectHandle, self).__init__(
          object_node,
          truncate,
          append_only,
        )

        self._start_streaming_thread()

    @convert_fs_exceptions
    def close(self):
        # FIXME: This is pretty tightly coupled to our base class's
        # implementation.

        # Note: We are overriding our base class.
        self.fobj.close()
        self._stop_streaming_thread()
        return SFTP_OK

    @convert_fs_exceptions
    def write(self, offset, data):
        if offset != self._next_offset:
            # Streaming assumptions have been violated.  Time to give up.
            logger.debug('%r != %r: no longer streaming', offset, self._next_offset)
            self._stop_streaming_thread()
            self.__class__ = WriteObjectHandle
            return self.write(offset, data)

        self._send(data)
        self._next_offset = offset + len(data)

        return super(OptimisticStreamingWriteObjectHandle, self).write(offset, data)


###


def node_attrs_to_sftp_attrs(node_attrs):
    sftp_attrs = SFTPAttributes()
    if node_attrs.st_size is not None:
        sftp_attrs.st_size = node_attrs.st_size
    if node_attrs.st_uid is not None:
        sftp_attrs.st_uid = node_attrs.st_uid
    if node_attrs.st_gid is not None:
        sftp_attrs.st_gid = node_attrs.st_gid
    if node_attrs.st_mode is not None:
        sftp_attrs.st_mode = node_attrs.st_mode
    if node_attrs.st_atime is not None:
        sftp_attrs.st_atime = node_attrs.st_atime
    if node_attrs.st_mtime is not None:
        sftp_attrs.st_mtime = node_attrs.st_mtime
    if node_attrs.metadata is not None:
        sftp_attrs.attr.update(node_attrs.metadata)
    if node_attrs.filename is not None:
        sftp_attrs.filename = node_attrs.filename
    return sftp_attrs


def sftp_attrs_to_node_attrs(sftp_attrs):
    return NodeAttributes(
      st_size = sftp_attrs.st_size,
      st_uid = sftp_attrs.st_uid,
      st_gid = sftp_attrs.st_gid,
      st_mode = sftp_attrs.st_mode,
      st_atime = sftp_attrs.st_atime,
      st_mtime = sftp_attrs.st_mtime,
      metadata = sftp_attrs.attr,
    )


class CloudFilesSFTPServerInterface(SFTPServerInterface):
    # This is a cloudfiles_sftpd.server_interface.CloudFilesServerInterface
    # instance.
    server = None
    # This is a paramiko.server.Transport instance.
    transport = None
    connpool = None

    access_policies = None
    auth_method_options = None

    session_info = None
    session_options = None
    cloudfiles_conn = None
    filesys = None

    def __init__(
      self,
      server,
      transport,
      connpool,
      access_policies,
      auth_method_options,
    ):
        super(CloudFilesSFTPServerInterface, self).__init__(server)

        self.server = server
        self.transport = transport
        self.connpool = connpool
        self.access_policies = access_policies
        self.auth_method_options = auth_method_options

    def session_started(self):
        self.session_info = {
          'auth_method': self.transport.server_object.active_auth_method,
          'username': self.transport.get_username(),
          'peer': self.transport.getpeername(),
          'container_locked': False,
        }
        access_manager = SessionPolicyAccessManager(
          self.access_policies,
          self.session_info,
        )
        self.cloudfiles_conn = self.connpool.get()

        self.session_options = self.auth_method_options.get(
          self.session_info['auth_method'].name,
          {},
        )

        root = self.session_options.get('root', '/')
        try:
            root = root.format(**self.session_info)
        except KeyError, e:
            raise ConfigurationError('invalid root format: %s (%s)' % (root, e))

        root = posixpath.normpath(root)
        if not root.startswith('/'):
            raise ConfigurationError('root must start with "/"')

        self.session_info['root'] = root

        _lock_container = self.session_options.get('lock_container', False)
        if _lock_container:
            session_container_name = self._get_session_container_name()
            if session_container_name:
                logger.info('locking container %s', session_container_name)
                # FIXME: How should we handle BusyError here?
                lock_container(self.cloudfiles_conn, session_container_name)
                self.session_info['container_locked'] = True
            else:
                logger.warning(
                  'no container locking: session not rooted within a container'
                )

        create_root = self.session_options.get('create_root', False)
        if create_root and (root != '/'):
            root_filesys = CloudFileSystem(self.cloudfiles_conn, AllowAccessManager())
            # FIXME: How should we handle errors checking the existence of
            # and/or creating the session root?  Right now we just let them
            # propagate.
            try:
                stat_result = root_filesys.lstat(root)
            except (NoSuchContainer, NoSuchObject):
                # FIXME: We should probably create parent directories if
                # necessary.
                root_filesys.mkdir(root)
            else:
                if not stat.S_ISDIR(stat_result.st_mode):
                    logger.error('root exists but is not a directory')
                    raise OperationError(errno.ENOTDIR)

        peer_addr, peer_port = self.session_info['peer']
        logger.info(
          'SFTP session started: user=%s, addr=%s, port=%s, root=%s, lock_container=%s, create_root=%s',
          self.session_info['username'],
          peer_addr,
          peer_port,
          root,
          _lock_container,
          create_root,
        )

        if root == '/':
            path_prefix = ''
        else:
            path_prefix = root
        if path_prefix.endswith('/'):
            raise AssertionError(path_prefix)

        self.filesys = CloudFileSystem(
          self.cloudfiles_conn,
          access_manager,
          path_prefix = path_prefix,
        )

    def session_ended(self):
        logger.info(
          'SFTP session ended: user=%s, addr=%s, port=%s',
          self.session_info['username'],
          *self.session_info['peer']
        )
        if self.session_info['container_locked']:
            session_container_name = self._get_session_container_name()
            logger.info('unlocking container %s', session_container_name)
            unlock_container(self.cloudfiles_conn, session_container_name)
        if self.filesys is not None:
            del self.filesys
        if self.cloudfiles_conn is not None:
            self.connpool.put(self.cloudfiles_conn)
            del self.cloudfiles_conn

    def _get_session_container_name(self):
        root = self.session_info['root']
        if root != '/':
            return root[1:].partition('/')[0]
        return None

    @convert_fs_exceptions
    def open(self, path, flags, attr):
        object_node = self.filesys.open(
          path,
          (flags & (
            os.O_CREAT
            | os.O_EXCL
            | os.O_RDONLY
            | os.O_WRONLY
            | os.O_RDWR
          )),
          sftp_attrs_to_node_attrs(attr),
        )
        return make_handle(object_node, flags)

    @convert_fs_exceptions
    def list_folder(self, path):
        return [node_attrs_to_sftp_attrs(entry) for entry in self.filesys.listdir(path)]

    @convert_fs_exceptions
    def stat(self, path):
        return node_attrs_to_sftp_attrs(self.filesys.stat(path))

    @convert_fs_exceptions
    def lstat(self, path):
        return node_attrs_to_sftp_attrs(self.filesys.lstat(path))

    @convert_fs_exceptions
    def remove(self, path):
        self.filesys.remove(path)
        return SFTP_OK

    @convert_fs_exceptions
    def rename(self, oldpath, newpath):
        self.filesys.rename(oldpath, newpath)
        return SFTP_OK

    @convert_fs_exceptions
    def mkdir(self, path, attr):
        self.filesys.mkdir(path, sftp_attrs_to_node_attrs(attr))
        return SFTP_OK

    @convert_fs_exceptions
    def rmdir(self, path):
        self.filesys.rmdir(path)
        return SFTP_OK

    @convert_fs_exceptions
    @logged(logger)
    def chattr(self, path, attr):
        self.filesys.chattr(path, sftp_attrs_to_node_attrs(attr))
        return SFTP_OK

    def canonicalize(self, path):
        return self.filesys.canonicalize(path)

    @convert_fs_exceptions
    def readlink(self, path):
        return self.filesys.readlink(path)

    @convert_fs_exceptions
    def symlink(self, target_path, path):
        self.filesys.symlink(target_path, path)
        return SFTP_OK
